﻿using System.IO;
using System.Runtime.Loader;
using Mono.Cecil;

var appname = Path.GetFileName (Environment.GetCommandLineArgs ()[0]);
if (args.Length != 2) {
    Console.WriteLine ($"Usage: dotnet {appname} INPUT OUTPUT");
    return;
}

var input   = args [0];
var output  = args [1];

var context = new AssemblyLoadContext ("input loader");
var assembly  = context.LoadFromAssemblyPath (Path.GetFullPath (input));
var MyType_type = assembly.GetType ("Example.MyType");

var assemblyDef = AssemblyDefinition.ReadAssembly (input);

var MyType_ref = assemblyDef.MainModule.ImportReference (MyType_type);

AddDelegateDeclaration (assemblyDef.MainModule, MyType_ref);

assemblyDef.Write (output);

static void AddDelegateDeclaration (ModuleDefinition module, TypeReference parameterType)
{
    var typeDef = CreateDelegateType (module, "MyDelegate", new [] {new ParameterDefinition ("type", default, parameterType)}, module.TypeSystem.Void);
    module.Types.Add (typeDef);
}

static TypeDefinition CreateDelegateType (ModuleDefinition module, string delegateName, IList<ParameterDefinition> parameters, TypeReference returnType)
{
    var delegateDef = new TypeDefinition (
            @namespace: "",
            name:       delegateName,
            attributes: TypeAttributes.Class | TypeAttributes.NotPublic | TypeAttributes.Sealed | TypeAttributes.AnsiClass | TypeAttributes.AutoClass
    );
    delegateDef.BaseType = module.ImportReference (typeof (MulticastDelegate));

    var delegateCtor = new MethodDefinition (
            name:       ".ctor",
            attributes: MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.SpecialName | MethodAttributes.RTSpecialName,
            returnType: module.TypeSystem.Void
    );
    delegateCtor.ImplAttributes     = MethodImplAttributes.Runtime | MethodImplAttributes.Managed;
    delegateCtor.Parameters.Add (new ParameterDefinition ("object", default, module.TypeSystem.Object));
    delegateCtor.Parameters.Add (new ParameterDefinition ("method", default, module.TypeSystem.IntPtr));
    delegateDef.Methods.Add (delegateCtor);

    var invoke = new MethodDefinition (
            name: "Invoke",
            attributes: MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.NewSlot | MethodAttributes.Virtual,
            returnType: returnType
    );
    invoke.ImplAttributes   = MethodImplAttributes.Runtime | MethodImplAttributes.Managed;
    foreach (var p in parameters) {
        invoke.Parameters.Add (new ParameterDefinition (p.Name, p.Attributes, p.ParameterType));
    }
    delegateDef.Methods.Add (invoke);

    return delegateDef;
}
